/*
 * Copyright (c) 2022 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/**
 * @file
 *   This file implements a generic bluetooth external radio 3w coexistence
 *   interface.
 *
 */

#include <stddef.h>
#include <stdint.h>

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/devicetree.h>
#include <nrfx.h>
#include <nrfx_ppi.h>
#include <nrfx_gpiote.h>

#include <mpsl_coex.h>

#if DT_NODE_HAS_STATUS(DT_PHANDLE(DT_NODELABEL(radio), coex), okay)
#define COEX_NODE DT_PHANDLE(DT_NODELABEL(radio), coex)
#else
#define COEX_NODE DT_INVALID_NODE
#error No enabled coex nodes registered in DTS.
#endif

#if !(DT_NODE_HAS_COMPAT(COEX_NODE, generic_radio_coex_three_wire))
#error Selected coex node is not compatible with generic-radio-coex-three-wire.
#endif

#if !IS_ENABLED(CONFIG_SOC_SERIES_NRF52X)
#error Bluetooth 3-wire coex is only supported on the nRF52 series
#endif

#define MPSL_COEX_BT_GPIO_POLARITY_GET(dt_property)                                                \
	((GPIO_ACTIVE_LOW & DT_GPIO_FLAGS(COEX_NODE, dt_property)) ? false : true)

static volatile bool coex_enabled;

static void coex_enable_callback(void)
{
	coex_enabled = true;
}

static const struct gpio_dt_spec req_spec = GPIO_DT_SPEC_GET(COEX_NODE, req_gpios);
static const struct gpio_dt_spec pri_spec = GPIO_DT_SPEC_GET(COEX_NODE, pri_dir_gpios);
static const struct gpio_dt_spec gra_spec = GPIO_DT_SPEC_GET(COEX_NODE, grant_gpios);

static int mpsl_cx_bt_interface_3wire_config_set(void)
{
	nrfx_err_t err = NRFX_SUCCESS;
	nrf_ppi_channel_t ppi_channel;
	mpsl_coex_if_t coex_if_bt;
	mpsl_coex_802152_3wire_gpiote_if_t *coex_if = &coex_if_bt.interfaces.coex_3wire_gpiote;
    int32_t ret;
    
    ret = gpio_pin_configure_dt(&req_spec, GPIO_OUTPUT_INACTIVE);
	if (ret != 0) {
		return ret;
	}

	ret = gpio_pin_configure_dt(&pri_spec, GPIO_OUTPUT_INACTIVE);
	if (ret != 0) {
		return ret;
	}

	ret = gpio_pin_configure_dt(&gra_spec, GPIO_INPUT | GPIO_PULL_UP);
	if (ret != 0) {
		return ret;
	}

	/* Allocate GPIOTE and PPI channels for the 3 lines */
    mpsl_coex_gpiote_cfg_t *gpiote_cfg[3] = {
        &(coex_if->grant_cfg),
        &(coex_if->priority_cfg),
        &(coex_if->request_cfg)
    };
    uint8_t gpio_pin[3] = {
        NRF_DT_GPIOS_TO_PSEL(COEX_NODE, grant_gpios),
        NRF_DT_GPIOS_TO_PSEL(COEX_NODE, pri_dir_gpios),
        NRF_DT_GPIOS_TO_PSEL(COEX_NODE, req_gpios),
    };
    uint8_t gpio_polar[3] = {
        MPSL_COEX_BT_GPIO_POLARITY_GET(grant_gpios),
        MPSL_COEX_BT_GPIO_POLARITY_GET(pri_dir_gpios),
        MPSL_COEX_BT_GPIO_POLARITY_GET(req_gpios),
    };

    for(int i = 0; i < 3; i++){
        if (nrfx_gpiote_channel_alloc(&(gpiote_cfg[i]->gpiote_ch_id)) != NRFX_SUCCESS) {
            return -ENOMEM;
        }
        if (nrfx_ppi_channel_alloc(&ppi_channel) != NRFX_SUCCESS) {
            return -ENOMEM;
        }
        gpiote_cfg[i]->gpio_pin = gpio_pin[i];
        gpiote_cfg[i]->active_high = gpio_polar[i];
        gpiote_cfg[i]->ppi_ch_id = ppi_channel;
    }

    /* Allocate additional PPI channel for setting the priority level at the correct time */
    /* Since Priority and Status are combined functional, it requires separate PPI channel to handle both */
    if (nrfx_ppi_channel_alloc(&ppi_channel) != NRFX_SUCCESS) {
        return -ENOMEM;
    }
    coex_if->additional_ppi_ch_id = ppi_channel;

    /* When the PRIORITY pin indicates the Status.*/
    /* TX is the active level */
    coex_if->is_rx_active_level = 0;

    /* Delay in us from the request pin is raised until the Priority and Status line shows the type of transaction (RX/TX). */
    coex_if->type_delay_us = 0;

    /* Delay in us from the Priority and Status line shows the type of transaction, to the radio starts its on-air activity. */
    coex_if->radio_delay_us = 50;

    /* Pointer to a timer instance. Timer should not be shared with any other functionality. */
    coex_if->p_timer_instance = NRF_TIMER2;


	/* Enable support for coex APIs (make sure it's not link-time optimized out) */
	mpsl_coex_support_802152_3wire_gpiote_if();

	/* Enable 1-wire coex implementation */
	coex_if_bt.if_id = MPSL_COEX_802152_3WIRE_GPIOTE_ID;
	err = (int)mpsl_coex_enable(&coex_if_bt, coex_enable_callback);
    __ASSERT(err, "Failed to configure finished");

	/** If enable API call is successful, wait for initialization to complete
	 *  by waiting for the supplied callback to be called.
	 */
	while (!err && !coex_enabled) {
		k_yield();
	}
    printk("3-wire coex enabled\n");
	return err;
}

static int mpsl_cx_init(void)
{

	return mpsl_cx_bt_interface_3wire_config_set();
}

SYS_INIT(mpsl_cx_init, APPLICATION, CONFIG_KERNEL_INIT_PRIORITY_DEVICE);